# Intelligent disassembler for Wang 3300 binary code.
# Jim Battle, 2006
#
# *) process user configuration file
# *) scan through the file marking addresses with non-legal instructions
# *) conservative scan.
#    scan memory for known instructions.  we start tracing from that point
#    only if the instruction doesn't care what the P0/P1 flag is, or if it
#    doesn't matter to the flow control (branch absolute).  all potential
#    flows out of that instruction are followed, and they inherit the page
#    status that came in to them.  we don't follow indirect jumps, but
#    indirect subroutine calls are followed on the grounds that those
#    locations are unlikely to be changed.  a stream isn't followed unless
#    the instruction in question hasn't been visited before or if the
#    page status would change as a result of the visit.
# *) more lenient scan.
#    this uses the same machinery as the conservative pass, but flags are set
#    allowing more speculative tracing, such as following through indirections
#    that aren't modified.  if we are later proven wrong, then a note is made.
# *) disassemble each location, tagged with all the gleaned information
# *) print the symbol table
#
# TODO:
# *) add mechanism for subroutine call to autocomment, so for example all
#    callers of INTSUB would get the comment "; form integer subscript in A"
#
# *) because setAttr checks for duplicates, you can't push two ";"
#    comment lines to one address
#
# *) for cases like: "LAI X'00'" where the imm is modified elsewhere,
#    don't print precomment mentioning that fact, instead list it as
#          LAI X'nn'
#    or something like that.  make it controlled by an option var.
#
# *) ideally the JST autocomments would not be inserted into the $attr
#    COMMENTS list, just into the rvinfo list.
#
# *) INDIR_OP isn't hooked up -- OPS always indirect.
#
# *) allow "DS n" specifier to declare uninitialized space
#
# *) allow generating output in a format suitable by use for the
#    emulator so it can report annotated lines as it steps.
#
# *) seek out FIXME comments

use strict;

# disassembly formatting control.  each line follows this format:
#  <address>: <1-4 data bytes> <mref> <tags> <label> <opcode> <operand> <comments>
my $DATA_COLUMN    =  7;   # which column data bytes begin
my $DATA_BYTES     =  2;   # number of bytes that can be printed on one row
my $MREF_COLUMN    =       # memory reference field
        $DATA_COLUMN + 2*$DATA_BYTES + 1;
my $TAG_COLUMN     =       # opcode tags
        $MREF_COLUMN + 10;
my $LABEL_COLUMN   =       # which column data bytes begin
        $TAG_COLUMN + 4;
my $OPCODE_COLUMN  =       # which column opcode begins
        $LABEL_COLUMN + 8;
my $ARG_COLUMN     =       # which column argument begins
        $OPCODE_COLUMN + 5;
my $COMMENT_COLUMN =       # which column comments begin
        $ARG_COLUMN + 9;

# this controls how wide to make the precomment fills
my $FILL_COLUMN = 80;

##########################################################################
# this table contains information about decoding the 3300 instructions.
# the encoded data in each string is described here:
#
#      i = expanded into 8b immediate as X'dd'
#      n = one's complemented and then treated like 'i' above
#      a = page 0/1 absolute reference, treated like 'i' for now
#      c = current page address relative, expanded as 16b X'dddd'
#      j = like c, but if the value is equal to *+2, don't print anything
#          used for optional jump files.  different from other types in that
#          it also affects the way the mnemonic is printed
#      s = (value of bits [1:0] + 1) are printed (1-4)
#          used for shift/rotate instructions
#      x = illegal opcode, and just X'dd' is printed
#      z = sign extend 8b immediate to 16b and print as X'dddd'

# page 98 (TABLE A.1) of the developers manual refers to these unknown ops:
#     STG  (SKIP IF GENERAL TRUE)
#     SFG  (SKIP IF GENERAL FALSE)
#     ICH  (INITIATE CHANNEL)
#     ECH  (END CHANNEL)
#
# SFG  is also referenced in the same manual on page
#
# STG and SFG appear on page 55 of wang3300.djvu from steve witham,
# but they aren't explained.

my @decodeTable = (
    "HLTj",       # 0x00
    "PARj",       # 0x01
    "Nj",         # 0x02   appendix c lists this as "INj"
    "DNj",        # 0x03
    "AZAj",       # 0x04
    "XZAj",       # 0x05
    "TSAj",       # 0x06
    "TZAj",       # 0x07
    "SBj",        # 0x08
    "SBCj",       # 0x09
    "SDj",        # 0x0A
    "LZAj",       # 0x0B
    "ONS i; f",   # 0x0C
    "OFS n; f",   # 0x0D
    "TASj",       # 0x0E
    "JZA",        # 0x0F

    "STA i",      # 0x10
    "SFA n",      # 0x11
    "SAA i",      # 0x12
    "SMA n",      # 0x13
    "STS i; f",   # 0x14
    "SFS n; f",   # 0x15
    "STG i",      # 0x16   not sure if this is valid
    "SFG n",      # 0x17   not sure if this is valid
    "AI i",       # 0x18
    "CI i",       # 0x19
    "LAI i",      # 0x1A
    "LZI i",      # 0x1B
    "DLI z",      # 0x1C
    "BAI i",      # 0x1D
    "BXI i",      # 0x1E
    "BOI i",      # 0x1F

    "SH s",       # 0x20
    "SHC s",      # 0x21
    "RT s",       # 0x22
    "RTC s",      # 0x23
    "JGT c",      # 0x24   aka "JNC c"
    "JNE c",      # 0x25   aka "JNZ c"
    "JLT c",      # 0x26
    "JEQ c",      # 0x27   aka "JZ c"
    "x",          # 0x28
    "x",          # 0x29
    "x",          # 0x2A
    "x",          # 0x2B
    "x",          # 0x2C
    "x",          # 0x2D
    "x",          # 0x2E
    "x",          # 0x2F

    "STI i",      # 0x30
    "SFI n",      # 0x31
    "AKIj",       # 0x32
    "DSIj",       # 0x33
    "ICH?",       # 0x34   initiate channel (not sure if this is right)
    "ECH?",       # 0x35   end channel      (not sure if this is right)
    "x",          # 0x36
    "x",          # 0x37
    "TIAj",       # 0x38
    "ONMj",       # 0x39
    "RDDj",       # 0x3A
    "WRDj",       # 0x3B
    "CIOj",       # 0x3C
    "x",          # 0x3D
    "x",          # 0x3E
    "x",          # 0x3F

    "INC a", "INC* a", "INC c", "INC* c",  # 0x40
    "JMP a", "JMP* a", "JMP c", "JMP* c",  # 0x44
    "BA a",  "BA* a",  "BA c",  "BA* c",   # 0x48
    "JST a", "JST* a", "JST c", "JST* c",  # 0x4C

    "BX a",  "BX a",   "BX c",  "BX c",    # 0x50
    "JEI a", "JEI* a", "JEI c", "JEI* c",  # 0x54
    "BO a",  "BO* a",  "BO c",  "BO* c",   # 0x58
    "x",     "x",      "x",     "x",       # 0x5C

    "x", "x", "x", "x", "x", "x", "x", "x",  # 0x60-0x67
    "x", "x", "x", "x", "x", "x", "x", "x",  # 0x68-0x6F

    "x", "x", "x", "x", "x", "x", "x", "x",  # 0x70-0x77
    "x", "x", "x", "x", "x", "x", "x", "x",  # 0x78-0x7F

    "ADD a", "ADD* a", "ADD c", "ADD* c", "ADD+ a", "ADD- a", "ADD+ c", "ADD- c",  # 0x80
    "AC a",  "AC* a",  "AC c",  "AC* c",  "AC+ a",  "AC- a",  "AC+ c",  "AC- c",   # 0x88
    "C a",   "C* a",   "C c",   "C* c",   "C+ a",   "C- a",   "C+ c",   "C- c",    # 0x90
    "AMC a", "AMC* a", "AMC c", "AMC* c", "AMC+ a", "AMC- a", "AMC+ c", "AMC- c",  # 0x98
    "DCM a", "DCM* a", "DCM c", "DCM* c", "DCM+ a", "DCM- a", "DCM+ c", "DCM- c",  # 0xA0
    "DAM a", "DAM* a", "DAM c", "DAM* c", "DAM+ a", "DAM- a", "DAM+ c", "DAM- c",  # 0xA8
    "DL a",  "DL* a",  "DL c",  "DL* c",  "DL+ a",  "DL- a",  "DL+ c",  "DL- c",   # 0xB0
    "DU a",  "DU* a",  "DU c",  "DU* c",  "DU+ a",  "DU- a",  "DU+ c",  "DU- c",   # 0xB8
    "LA a",  "LA* a",  "LA c",  "LA* c",  "LA+ a",  "LA- a",  "LA+ c",  "LA- c",   # 0xC0
    "UA a",  "UA* a",  "UA c",  "UA* c",  "UA+ a",  "UA- a",  "UA+ c",  "UA- c",   # 0xC8
    "LZ a",  "LZ* a",  "LZ c",  "LZ* c",  "LZ+ a",  "LZ- a",  "LZ+ c",  "LZ- c",   # 0xD0
    "UZ a",  "UZ* a",  "UZ c",  "UZ* c",  "UZ+ a",  "UZ- a",  "UZ+ c",  "UZ- c",   # 0xD8
    "XMA a", "XMA* a", "XMA c", "XMA* c", "XMA+ a", "XMA- a", "XMA+ c", "XMA- c",  # 0xE0
    "UAH a", "UAH* a", "UAH c", "UAH* c", "UAH+ a", "UAH- a", "UAH+ c", "UAH- c",  # 0xE8
    "x",     "x",      "x",     "x",      "x",      "x",      "x",      "x",       # 0xF0-0xF7
    "x",     "x",      "x",     "x",      "x",      "x",      "x",      "x",       # 0xF8-0xFF
);

##########################################################################
# data structures for the disassembler follow, along with some
# utility functions for mucking with it

# this is set to different levels to enable debugging messages
my $dbg;

# these arrays have one value per memory location in 0x0000 to 0xFFFF
my @mem;        # memory byte value or undef if not declared
my @attr;       # attribute hash, one for each possible address

# this is a hash of options in effect for the current pass.
#     INDIR_JST  -- follow indirect JSTs
#     INDIR_JMP  -- follow indirect JMP and JEI
my %MarkOpt;

# this tracks which addresses have been inspected on this pass
my @Marked;

# this global counts how many locations have been updated in this pass
my $MarkChanges;

# tracks which pass we are running
my $MarkPass = 0;

# each location has an attribute hash; attributes are
#    LABEL        -- user supplied label for given address
#    PRECOMMENTS  -- [ user supplied comment lines that appear before address ]
#    COMMENTS     -- [ user supplied comment lines ]
#    IFORMAT      -- user requested immediate # format (C=char, B=binary, D=signed decimal, U=unsigned decimal)
#    KNOWN_DAC    -- user said it was a DAC address
#    KNOWN_DC     -- user said it was a data address
#    KNOWN_DC_LEN -- length of DC sequence of bytes
#    KNOWN_P0     -- user said this instruction is known to have P state zero
#    KNOWN_P1     -- user said this instruction is known to have P state one
#    LEGAL_INST   -- defined opcode and user didn't say it was data
#    INFR_INST    -- inferred instruction
#    BRANCH_SRC   -- [ list of branch sources ]
#    JST_SRC      -- [ list of JST sources ]
#    RD_SRC       -- [ list of locations that read this one as byte ]
#    RD2_SRC      -- [ list of locations that read this one as word ]
#    WR_SRC       -- [ list of locations that write this one as byte ]
#    WR2_SRC      -- [ list of locations that write this one as word ]
#    RMW_SRC      -- [ list of locations that rmw this one as byte ]
#    RMW2_SRC     -- [ list of locations that rmw this one as word ]
#    PTR_SRC      -- [ list of locations that use this as indirect ptr ]
#    P_STATE      -- contains state of the status register P flag;
#                    1 if is true; 0 it is is false; empty if unknown.
#                    it can contain both if two instruction traces pass
#                    by a point with differering P settings.

{
    # tags that are associated with scalar states
    my %scalarTag = map {($_,1)}
            qw ( KNOWN_INST KNOWN_DAC KNOWN_DC KNOWN_DC_LEN KNOWN_P0 KNOWN_P1
                 P_STATE IFORMAT LABEL INFR_INST LEGAL_INST JST_INFO
                 BRANCH MEMOP PCHANGE
               );

    # tags that are associated with lists
    my %listTag = map {($_,1)}
            qw ( PRECOMMENTS COMMENTS
                 BRANCH_SRC JST_SRC RD_SRC RD2_SRC WR_SRC WR2_SRC
                 RMW_SRC RMW2_SRC PTR_SRC );

    # for tags holding lists, push the value on the list if it isn't a duplicate.
    # for non-list tags, set the attribute to the supplied value.
    sub setAttr
    {
        my $addr = shift;
        my $tag  = shift;
        my $val  = shift;

        if ($scalarTag{$tag}) {
            $val = 1 if !defined($val);  # useful to default booleans to true
            if ($attr[$addr]{$tag} ne $val) {
                printf "Change: %04X tag changed from '%s' to '%s'\n",
                        $addr, $attr[$addr]{$tag}, $val
                    if ($dbg > 3);
                $attr[$addr]{$tag} = $val;
                $MarkChanges++;
            }
        } elsif ($listTag{$tag}) {
            # check that value is pushed only once since subsequent passes
            # might push attribute again.
            # FIXME: use a hash to speed this up?
            #        have separate listTag and hashTag types
            if (!grep { $_ eq $val } @{$attr[$addr]{$tag}}) {
                push @{$attr[$addr]{$tag}}, $val;
                printf "Change: %04X tag pushed '%s'\n", $addr, $val
                    if ($dbg > 3);
                $MarkChanges++;
            }
        } else {
            die "Bad tag '$tag'";
        }
    }

    # return true if the address has any of the tags in the associated list.
    # optionally, the 2nd param can be a single string tag instead of a list
    sub hasAttr
    {
        my ($addr,$rtags) = @_;

        return exists($attr[$addr]{$rtags})
            if !ref($rtags);

        foreach my $tag (@$rtags) {
            return 1 if exists($attr[$addr]{$tag});
        }

        return 0;
    }

    # for non-lists, return the value associated with the tag.
    # for lists, return the ref to the list of values associated with the supplied tag.
    sub getAttr
    {
        my ($addr,$tag) = @_;
        die "Bad tag '$tag'"
            if (!$scalarTag{$tag} and !$listTag{$tag});
        return $attr[$addr]{$tag} if (exists $attr[$addr]{$tag});
        return undef;
    }
}


# return 1 if address pair both contain valid bytes
sub valid16
{
    my $addr = shift;
    return defined($addr) && (($addr & 1) == 0) &&
           defined($mem[$addr]) && defined($mem[$addr+1]);
}


# return 16b value at specified address
sub get16
{
    my $addr = shift;
    return 256 * $mem[$addr] + $mem[$addr+1];
}


my %allLabels = ();
sub AddUserLabel
{
    my($addr,$label) = @_;
    if ($label) {
        if ($allLabels{$label}) {
            die "Label '$label' declared twice";
        }
        $allLabels{$label} = 1; # mark it as seen
        if (&hasAttr($addr,"LABEL")) {
            printf "Warning: label overrride at address 0x%04X ('%s' replaces '%s')\n",
                $addr, $label, &getAttr($addr,"LABEL");
        }
        setAttr($addr, "LABEL", $label);
    }
}

sub AddUserDC
{
    my($addr,$count) = @_;
# FIXME: punt for now and just mark each byte as a single DC
#    setAttr($addr, "KNOWN_DC_LEN", $count);
    for(my $i=$addr; $i<$addr+$count; $i++) {
        setAttr($i, "KNOWN_DC");
        setAttr($i, "KNOWN_DC_LEN", 1);
        #printf "Adding address 0x%04X as type DC\n", $i;
    }
}

##########################################################################
# simple one line disassembler
# given a 16b instruction, return a string with a disassembled version

# return the 16b address implied by the page 0/1 absolute addressing.
# whether page 0 or page 1 is used is only as accurate as the determination
# of when the P status bit is 0 or 1.  hopefully it is easily determined.
sub AbsAddr
{
    my $addr = shift;
    my $imm = $mem[$addr+1];
    return (&getAttr($addr,"P_STATE") =~ m/1/) ? ($imm + 0x100) :  $imm;
}


# return a string with a sorted list of addresses that refer to the
# given address ($addr) of a given type ($tag)
sub referer_list
{
    my($addr,$tag,$lbl) = @_;
    return "" if !&hasAttr($addr,$tag);
    return "; $lbl\[" . join(",", map { sprintf "%04X", $_ }
                                  sort { $a <=> $b }
                                  @{&getAttr($addr, $tag)}) . "]";
}

# return a string containing a 16b absolute address.
# use a label if one matches.  add a comment to the address showing what
# the label value really is.
sub addr_label
{
    my $addr = shift;
    my $label;

    if (($addr & 1) && !&hasAttr($addr,"LABEL") && IsValidInst($addr-1)) {
        # if this is referring to an odd byte of an instruction and
        # it doesn't have its own label, make a relative label
        $label = &getAttr($addr-1,"LABEL") . "+1"
              if &hasAttr($addr-1,"LABEL");
    } else {
        $label = &getAttr($addr,"LABEL");
    }
    return $label || sprintf("X'%04X'", $addr);
}

# return a string formatted according to the optional IFORMAT.
# the default is hex (X).
sub fmti8
{
    my $addr = shift;
    my $imm  = shift;
    my $fmt = &getAttr($addr, "IFORMAT");
    if ($fmt eq "C") {
        return sprintf("C'%c'", $imm)
            if ($imm >= ord(" ") and $imm <= ord("~"));
        # else fall through
    } elsif ($fmt eq "D") {
        # signed decimal
        $imm = ($imm-256) if ($imm >= 128);
        return sprintf("%d", $imm);
    } elsif ($fmt eq "U") {
        # unsigned decimal
        return sprintf("%d", $imm);
    } elsif ($fmt eq "B") {
        # binary
        my $s = "";
        for(my $n=128; $n>0; $n >>= 1) {
            $s .= ($imm & $n) ? "1" : "0";
        }
        return "B'$s'";
    }
    # else hex
    return sprintf "X'%02X'", $imm;
}

# return a string formatted according to the optional IFORMAT
# we don't respect the C (char) format.
# the default is signed decimal
sub fmti16
{
    my $addr = shift;
    my $imm  = shift;
    my $fmt = &getAttr($addr, "IFORMAT");
    if ($fmt eq "X") {
        # hex
        return sprintf "X'%04X'", $imm;
    } elsif ($fmt eq "U") {
        # unsigned decimal
        return sprintf("%d", $imm);
    } elsif ($fmt eq "B") {
        # binary
        my $s = "";
        for(my $n=32768; $n>0; $n >>= 1) {
            $s .= ($imm & $n) ? "1" : "0";
        }
        return "B'$s'";
    }
    # else signed decimal
    $imm = ($imm-65536) if ($imm >= 32768);
    return sprintf("%d", $imm);
}

# disassemble the information asssociated with one address.
# this is returned as a hash containing these elements:
#       BYTES: integer - length of 
#        MREF: memory address reference
#              undef if not meaningful; "" if unknown; integer if known
#        IREF: indirect memory address reference
#              undef if not meaningful; "" if unknown; integer if known
#        TAGS: string giving some terse information
#                 I=user specified instruction
#                 i=inferred instruction
#                 P=page 1 is in effect
#       LABEL: string
#      OPCODE: string
#     OPERAND: string
# PRECOMMENTS: reference to list
#    COMMENTS: reference to list
#        REFS: string -- locations referring to this address

sub dasm3300_core
{
    my $addr = shift;   # input
    my %rv;             # return value hash

    my $sflags   = "7PCZDV10";
    my $instr    = &get16($addr);
    my $op       = ($instr >> 8) & 0xFF;
    my $imm      = ($instr >> 0) & 0xFF;
    my $pageaddr = ($addr & 0xFF00) | $imm;

    # tag the line with some diagnostic information
    my $odd = ($addr & 1);
    if (!$odd) {
        if (&hasAttr($addr,"KNOWN_INST")) {
            $rv{TAGS} = "I";
        } elsif (&hasAttr($addr,"INFR_INST")) {
            $rv{TAGS} = "i";
        }
        if (IsValidInst($addr)) {
            # report known P state as true (P), false (0), or both (B);
            # else report P state as true (p), false (""), or both (b)
            my $idx1 = (&hasAttr($addr,"KNOWN_P1") ? 2 : 0)
                     + (&hasAttr($addr,"KNOWN_P0") ? 1 : 0);
            my $idx1c = substr(".0PB", $idx1, 1);
            my $ps = &getAttr($addr,"P_STATE");
            my $idx2 = (($ps =~ m/1/) ? 2 : 0)
                     + (($ps =~ m/0/) ? 1 : 0);
            my $idx2c = substr(".opb", $idx2, 1);
            $idx2c =~ tr/o//d;  # traced P=0 is common, so make it default
            $rv{TAGS} .= ($idx1) ? $idx1c : $idx2c;
        }
    }

    # indicates if this location is accessed as a byte pair
    # (it may also be accessed as individual bytes too)
    my $pair = IsValidInst($addr) ||
               (!($addr & 1) &&
                  &hasAttr($addr,["KNOWN_DAC", "PTR_SRC", "RD2_SRC",
                                  "WR2_SRC", "RMW2_SRC"]));

    # look for signs of misunderstanding and flag them
    # case 1: someone modifying the return address of a subroutine.  suppress
    #         the warning if the user has tagged this subroutine with JST.
    my $case1 = !($addr & 1) && &IsWritten($addr) && &hasAttr($addr,"JST_SRC") &&
                !&hasAttr($addr,"JST_INFO");
    my $case2 = &IsValidInst($addr) && &IsWritten($addr);
    # location that used as a DAC but code also reaches it
    my $case3 = &IsValidInst($addr) && &hasAttr($addr,"PTR_SRC");
    # location is referenced but isn't in binary image
    my $case4 = &Referenced($addr) && (!defined $mem[$addr] || !defined $mem[$addr+1]);
    # location uses absolute address mode but P_STATE is 01
    my $case5 = ($decodeTable[$mem[$addr]] =~ m/a/) &&  # absolute address
                (&getAttr($addr,"P_STATE") eq "01");

    $rv{TAGS} .= "?" if ($case5);

    # check for address label
    if (&hasAttr($addr,"LABEL")) {
        $rv{LABEL} = &getAttr($addr,"LABEL");
    } elsif ( &hasAttr($addr, ["BRANCH_SRC", "JST_SRC", "PTR_SRC",
                               "RD2_SRC", "WR2_SRC", "RMW2_SRC",
                               "RD_SRC", "WR_SRC", "RMW_SRC"]) ) {
        # no user supplied label, so make our own
        $rv{LABEL} = sprintf("L%04X", $addr);
    }

    # get any references to this address
    $rv{REFS} = &GetRefs($addr);
    if ($pair) {
        my $more = &GetRefs($addr+1);
        if ($more) {
            my $oddaddr = sprintf("%04X", $addr+1);
            $more =~ s/(\w+\[)/$oddaddr:\1/g;
            $rv{REFS} .= $more;
        }
    }

    # get user comments associated with this address
    $rv{PRECOMMENTS} = &getAttr($addr,"PRECOMMENTS");
    # autocomment any oddities that have been detected
    if (1) {
        my @msgs;
        push @msgs, "the return address gets modified"      if $case1;
        push @msgs, "the next instruction gets modified"    if $case2;
        push @msgs, "the next instruction also used as DAC" if $case3;
        push @msgs, "the next location referenced but not loaded" if $case4;
        push @{$rv{PRECOMMENTS}}, map { "; " . $_ } @msgs;
    }

    # autocomment subroutines if user hasn't done so
    my $rsubhash = &getSubroutineStatus($addr);
    if ($$rsubhash{SUB_OK} && !&hasAttr($addr, ["JST_INFO", "PRECOMMENTS"])) {
        my $leaf = ($$rsubhash{SUB_LEAF}) ? " leaf" : "";
        unshift @{$rv{PRECOMMENTS}},
            ";",
            "; ========== autodetected$leaf subroutine ==========";
    }

    $rv{COMMENTS} = &getAttr($addr,"COMMENTS");
    my $autodac = ($rv{REFS} eq "") && ($instr == $addr);

    if (  &hasAttr($addr,"KNOWN_DAC") || $autodac ||
         ((&hasAttr($addr,["PTR_SRC","JST_SRC"])) && !IsValidInst($addr)) ) {
        $rv{BYTES} = 2;
        $rv{OPCODE} = "DAC";
        $rv{OPERAND} = ($instr == $addr) ? "*" : &addr_label($instr);
        push @{$rv{COMMENTS}}, "; autodetected" if ($autodac);
        return \%rv;
    }

    # check user specified DC
    if (&hasAttr($addr,"KNOWN_DC")) {
        $rv{BYTES} = &getAttr($addr,"KNOWN_DC_LEN");
#printf "DC at %04X, length = %d bytes\n", $addr, $rv{BYTES};
        $rv{OPCODE} = "DC";
        for(my $a=$addr; $a<$addr+$rv{BYTES}; $a++) {
            $rv{OPERAND} .= sprintf("X'%02X',", $mem[$a]);
        }
        $rv{OPERAND} =~ s/,$//;  # remove trailing separator
        return \%rv;
    }

    # check for read pair or write pair to location
    if (!&IsValidInst($addr) && &hasAttr($addr,["RD2_SRC","WR2_SRC","RMW2_SRC"])) {
        $rv{BYTES} = 2;
        $rv{OPCODE} = "DC";
        $rv{OPERAND} = sprintf("X'%04X'", $instr);
        return \%rv;
    }

    # check for individual read or writes
    if (!IsValidInst($addr) &&
            ($odd || &hasAttr($addr,["RD_SRC","WR_SRC","RMW_SRC"])) ) {
        # FIXME: we could check for a run of DC bytes that aren't interrupted
        #        by a location with a label and return them all at once.
        $rv{BYTES} = 1;
        $rv{OPCODE} = "DC";
        $rv{OPERAND} = sprintf("X'%02X'", $mem[$addr]);
        return \%rv;
    }

    # check for illegal instruction
    my $dec = $decodeTable[$op];
    if ($dec eq "x") {
        $rv{BYTES} = 2;
        $rv{OPCODE} = "DC";
        $rv{OPERAND} = sprintf("X'%04X'", $instr);
        push @{$rv{COMMENTS}}, "; bad op";
        return \%rv;
    }

    die "bad op string" if ($dec !~ m/^(\S+)\s*([^;]*)(;.*)*/);
    my ($opcode,$operand,$cmnts) = ($1,$2,$3);

    $rv{OPCODE} = $opcode;
    if ($opcode =~ "j") {
        # current page relative address, suppressed
        if ($pageaddr != $addr+2) {
            $rv{OPCODE} =~ tr/j/J/;
            $rv{OPERAND} = ($addr == $pageaddr) ? "*"
                                                : &addr_label($pageaddr);
        } else {
            $rv{OPCODE} =~ tr/j//d;
        }
    }

    if ($operand eq "a") {
        # absolute page 0/1 address
        my $absaddr  = &AbsAddr($addr);
        if (&PageKnown($addr)) {
            $rv{OPERAND} = &addr_label($absaddr);
        } else {
            $rv{OPERAND} = sprintf("L0?%02X", $imm);
        }
    }
    elsif ($operand eq "i") {
        # 8b immediate
        $rv{OPERAND} = &fmti8($addr, $imm);
    }
    elsif ($operand eq "n") {
        # 8b immediate, complemented
        $imm = (~$imm) & 0xFF;
        $rv{OPERAND} = &fmti8($addr, $imm);
    }
    elsif ($operand eq "c") {
        # current page relative address
        $rv{OPERAND} = &addr_label($pageaddr);
    }
    elsif ($operand eq "s") {
        # shift count
        $rv{OPERAND} = sprintf("%d", ($imm & 3)+1);
    }
    elsif ($operand eq "z") {
        # 8b immediate sign extended to 16b
        my $imm16 = $imm
                  | (($imm & 0x80) ?  0xFF00 : 0x0000);
        $rv{OPERAND} = &fmti16($addr, $imm16);
    } elsif ($operand) {
        die "address $addr operand was '$operand'";
    }

    if (&needsMref($addr)) {
        my $mref = &getMref($addr);
        $rv{MREF} = defined($mref) ? $mref : "";
    }

    if (&needsIref($addr)) {
        if ($rv{MREF} eq "") {
            $rv{IREF} = "";
        } else {
            my $iref = &getIref($addr);
            $rv{IREF} = defined($iref) ? $iref : "";
        }
    }

    # if user has commented line, suppress some autocomments
    my $hasUserComments = &hasAttr($addr, "COMMENTS");

    # autocomment status flags
    if (1) {
        if ($cmnts =~ "f") {
            # print immediate as decoded flags
            my $flags = "";
            for(my $k = 7; $k >= 0; $k--) {
                if (($imm >> $k) & 1) {
                    $flags .= substr($sflags,7-$k,1);
                }
            }
            push @{$rv{COMMENTS}}, "; S.$flags";
        } elsif ($cmnts) {
            die "address $addr cmnts was '$cmnts'";
        }
    }

    # autocomment return jumps if the user hasn't commented line
    if (1) {
        if (($opcode eq "JMP*") and !$hasUserComments) {
            my $tgtaddr = &EffAddr($addr);   # current or absolute
            if (&hasAttr($tgtaddr,"JST_SRC")) {
                push @{$rv{COMMENTS}}, "; return";
            }
        }
    }

    # autocomment the loaded DL value if it is constant
    if (1) {
        if ($dec =~ m/^DL[ *]/) {
            my $tgtaddr = ($dec =~ m/\*/) ? &getIref($addr) : &getMref($addr);
            if (defined($tgtaddr) && !IsModified($tgtaddr)) {
                my $val = get16($tgtaddr);
                push @{$rv{COMMENTS}}, sprintf("; =X'%04X'", $val);
            }
        }
    }

    # autocomment the indirect target address if it is constant
    # this is now handled with the IREF field
    if (0) {
        if ($dec =~ m/\*/) {
            my $tgtaddr = &EffAddr($addr);   # current or absolute
            if (!IsModified($tgtaddr)) {
                my $ea = get16($tgtaddr);
                push @{$rv{COMMENTS}}, sprintf("; @%04X", $ea);
            }
        }
    }

    $rv{BYTES} = 2;

    return \%rv;
}


# disasssemble the line at the specified address and return it as a string.
# it may be more than one line, in which case they are separated by "\n".
sub dasm3300
{
    my $addr = shift;
    my $rhash = &dasm3300_core($addr);
    my $did_refs = 0;  # we've emitted REFS already

    # precomments appear on a line themselves before the addressed location
    my $precomments;
    if (0) {
        $precomments = join "", map { (" " x ($LABEL_COLUMN-1)) . $_ . "\n" }
                                      @{$$rhash{PRECOMMENTS}};
    } else {
        # primitive markup:
        # if a line contains only "--" or "==", then fill with that symbol
        # if a line begins and ends in "--" or "==", then center text and fill
        # leading spaces are honored and ignore trailing spaces are ignored.
        foreach my $pc (@{$$rhash{PRECOMMENTS}}) {
            my $ignore = ($pc =~ /autodetected/);
            my $msg;
            if (($pc =~ m/^ (;\s*)           # $1: optional significant spaces
                            ([#=-]){2,}      # $2: break character
                            (?:\s*)          #     ignored spaces
                        $/x) && !$ignore) {
                my $pad = $FILL_COLUMN - $LABEL_COLUMN - length($1);
                $msg = $1 . ($2 x $pad);
            } elsif (($pc =~ m/^ (;\s*)      # $1: optional significant spaces
                                 ([#=-]){2,} # $2: break character
                                 (?:\s*)     #     ignored spaces
                                 (.+?)       # $3: comment
                                 (?:\s*)     #     ignored spaces
                                 \2{2,}      #     repeated break character
                                 (?:\s*)     #     ignored spaces
                            $/x) && !$ignore) {
                my $fixed = $LABEL_COLUMN + length($1) + length($3) + 2;
                my $pad1 = int(($FILL_COLUMN - $fixed)/2);
                my $pad2 = $FILL_COLUMN - $fixed - $pad1;
                $msg = $1 . ($2 x $pad1) . " $3 " . ($2 x $pad2);
            } else {
                $msg .= $pc;
            }
            $precomments .= (" " x ($LABEL_COLUMN-1)) . $msg . "\n";
        }
    }

    my $hexaddr = sprintf("%04X:", $addr);


    # we put the label on its own line if either the label is wide.
    # append any REFS.
    my $label;
    if ($$rhash{LABEL} && (length($$rhash{LABEL}) > 7)) {
        $label  = $hexaddr;
        pad(\$label, $LABEL_COLUMN);
        $label .= $$rhash{LABEL};
        if ($$rhash{REFS}) {
            pad(\$label, $COMMENT_COLUMN);
            $label .= $$rhash{REFS};
        }
        $label .= "\n";
        $did_refs = 1;
    }

    my $retstr = $hexaddr;

    # FIXME: at the moment, dasm3300_core only returns a max BYTES of 2.
    #        if we allow it to be more, the following code will need to
    #        split the report into multiple lines.
    pad(\$retstr, $DATA_COLUMN);
    for(my $a=$addr; $a <$addr + $$rhash{BYTES}; $a++) {
        $retstr .= sprintf("%02X", $mem[$a]);
    }

    pad(\$retstr, $MREF_COLUMN);
    if (defined($$rhash{MREF})) {
        $retstr .= ($$rhash{MREF} eq "") ? "****"
                                         : sprintf("%04X", $$rhash{MREF});
        if (defined($$rhash{IREF})) {
            $retstr .= ($$rhash{IREF} eq "") ? ":****"
                                             : sprintf(":%04X", $$rhash{IREF});
        }
    }

    pad(\$retstr, $TAG_COLUMN);
    $retstr .= $$rhash{TAGS};

    if ($$rhash{LABEL} && !$label) {
        pad(\$retstr, $LABEL_COLUMN);
        $retstr .= $$rhash{LABEL};
    }

    pad(\$retstr, $OPCODE_COLUMN);
    $retstr .= $$rhash{OPCODE};

    pad(\$retstr, $ARG_COLUMN);
    $retstr .= $$rhash{OPERAND};

    if ($$rhash{COMMENTS}) {
        my @cmnts = @{$$rhash{COMMENTS}};
        my $pad = " " x ($COMMENT_COLUMN-1);
        pad(\$retstr, $COMMENT_COLUMN);
        $retstr .= $cmnts[0];
        for my $i (1 .. $#cmnts) {
            $retstr .= "\n" . $pad . $cmnts[$i];
        }
    } elsif (!$did_refs && $$rhash{REFS}) {
        pad(\$retstr, $COMMENT_COLUMN);
        $retstr .= $$rhash{REFS};
        $did_refs = 1;
    }
    $retstr .= "\n";

    my $refstr;
    if (!$did_refs && $$rhash{REFS}) {
        pad(\$refstr, $COMMENT_COLUMN);
        $refstr .= $$rhash{REFS} . "\n";
    }

    (my $final = $precomments . $label . $retstr . $refstr)
            =~ s/\s+$//;  # kill trailing spaces

    return [ $$rhash{BYTES}, $final ];
}

##########################################################################
# read one of the .hex files (NOT intel hex) that steve provided and
# do line-by-line disassembly.  the parsing here is pretty rigid and
# matches exactly what steve provided.

sub readWithamHex
{
    my $filename = shift;
    local $.;  # line number of file

    open(INFILE, "<$filename") || die "Couldn't open hex file '$filename'";

    while(<INFILE>) {
        chomp;
        if (/^\s*$/) { next; }  # skip blank lines

        my @words = split(/ +/);

        # get address
        if ($words[0] !~ /^(....):$/) {
            die "Bad address format, line $.";
        }
        my $addr = hex($1);
        # printf "line $., address 0x%04X, %d bytes\n", $addr, $#words;
        shift @words;

        # process list of bytes
        foreach my $w (@words) {
            $mem[$addr++] = hex($w);
        }
    }

    close INFILE;

    return 1;
}

##########################################################################
# Read the configuration file and parse out the data.
# The configuration file consists of a set of independent lines.
#
# + Blank lines are ignored.
#
# + Comments start with '#' and are ignored.
#
# + The generic format is as follows:
#      [<address>] [<format>] [<label>] [<directive>] [<args>]
#                                         [; user comment] [# hidden comment]
#
#   The <address> is a four digit hex number and must start with "."
#   (otherwise some addresses look like labels, and leading 0x gets tiresome
#   to type).  Examples are ".0123" and ".ACE4".  Upper and lower case hex
#   are allowed.  If no address is specified, the most recently specified
#   address is assumed.
#
#   The <format> string specifies how immediate operations should
#   appear in the listing.  The format is one of
#       B' -> binary
#       C' -> ascii character
#       D' -> signed decimal
#       U' -> unsigned decimal
#       X' -> hex
#   <format> is accepted with all commands, but it only makes sense with
#   some of them.  The format specifier can actually be anywhere before
#   the comment, not just in the location shown above.
#
#   User comments begin with first ";" in the line and include all text to
#   the end of line.  The comment appears in the listing; more than one
#   comment can be applied to the same address by having more than one
#   comment line specifying the same address.
#
#   If the line contains no user comment, there can be a config file comment
#   that doesn't appear in the listing.  This type of comment starts with
#   a "#" and goes to the end of line.
#
#   The <directive>s are:
#
#         I[0|1]  -- specifies this is a known instruction, optionally
#                    specifying the state of the "P" status bit.
#
#         JST [@@ ("param"|"ret")(:<comment>)*]+
#                 -- this specifies semantics around a subroutine
#                    this command allows specifying that a number of words
#                    after the call to the routine are to be marked with
#                    a comment and/or treated specially.  this is needed
#                    because the return address often gets diddled.
#
#         DC                -- mark one byte as being data
#         DC [+<bytecount>] -- mark a number of bytes in a row as being data;
#                              bytecount is decimal
#
#         DAC  -- specify byte pair as being an address pointer
#
#         H <filename> -- read in the Witham formatted hex file;
#                         the address, if supplied, is ignored.

sub ProcessConfig
{
    my $filename = shift;

    open(CFGFILE, "<$filename") or die "Couldn't open config file '$filename'";

    my $ptrnLabel = "[A-Za-z\$#@][A-Za-z0-9\$#@]*";  # label name
    my $ptrnAddr  = "0x[0-9a-fA-F]+|[0-9]+";  # hex or decimal

    my $addr = 0x0000;  # tracks most recent address

    while (<CFGFILE>) {

        chomp;
        #print "line1:'$_'\n";
        s/^ \s* (.*?) \s* $/$1/x;   # kill leading/trailing spaces
        next if (/^ \s* \#/x);      # leading "#"

        # extract optional user comment
        # it begins at the first semicolon on the line
        my $comment;
        if (/^ ([^;]*?) \s* (;.*) $/x) {
            $comment = $2;
            $_ = $1;
        }

        # strip hidden comment "# ..." and preceding spaces
        s/\s* \# .* $//x;
        #print "line2:'$_'\n";

        # <address>
        if (/^ [.] ([0-9A-Fa-f]{4}) \b \s* (.*) $/x) {
            $addr = hex($1);
            checkAddr($addr,$.);
            ##printf "Addr: %04X\n", $addr;
            $_ = $2;
        }
        #printf "%04X: comment='$comment'\n", $addr if defined($comment);

        # if a comment was specified, add it
        # if the comment starts as ";-", make it a precomment
        if (defined($comment)) {
            if ($comment =~ /^;-/) {
                $comment =~ s/-//; # just the first one
                setAttr($addr, "PRECOMMENTS", $comment);
            } else {
                setAttr($addr, "COMMENTS", $comment);
            }
        }

        # read in hex file -- this has to be done early since
        # it is different than the other commands
        # H <filename>
        if (m/^ H \s+ \" ([^\"]+) \" $/x) {
            my $hexname = $1;
            #printf "H:filename='%s'\n", $hexname;
            &readWithamHex($hexname);
            next;
        }

        # <format> specifier -- look for it anywhere
        if (/^ (.*?) \s* \b ([BCDUX])' \s* (.*) $/x) {
            #printf "%04X: iformat='$2'\n", $addr;
            setAttr($addr, "IFORMAT", $2);
            $_ = $1 . $3;
        }

        # if there is a label, it comes after the address
        if (/^ ($ptrnLabel) \b \s* (.*) $/xo) {
            my ($label,$rest) = ($1,$2);
            if ($label !~ /^ (DAC|DC|H|I|I0|I1|JST) $/x) {
                #printf "%04X: label='$label'\n", $addr;
                AddUserLabel($addr, $label);
                $_ = $rest;
            }
        }
        #print "line3:'$_'\n";

        # I[01] instruction specifier
        if (/^ I([01])? $/x) {
            my $page = $1;
            checkEvenAddr($addr,$.);
            setAttr($addr, "KNOWN_INST");
            setAttr($addr, "KNOWN_P$page") if ($page ne "");
            next;
        }

        # define subroutine semantics
        #   JST [@@ ("param"|"ret")(:<comment>)*]*
        # if there are no args, it means the JST doesn't return
        if (m/^JST  \b \s* (.*) $/x) {
            my @info = split m/\s* @@ \s+/x, $1;
            checkEvenAddr($addr,$.);
            if (@info) {
                die "Bad JST declaration"
                    if grep {$_ !~ m/^(param|ret)(:.*)?$/i} @info;
            }
            #printf "addr=%04X: JST info=%s\n", $addr, join(" @@ ",@info);
            setAttr($addr, "JST_INFO", \@info); # the list might be empty
            next;
        }

        # address data
        if ($_ eq "DAC") {
            checkEvenAddr($addr,$.);
            setAttr($addr, "KNOWN_DAC");
            next;
        }

        # define constant location(s)
        # DC [+<byte count>]
        if (m/^ DC \b \s* ([+]([0-9]+))? $/x) {
            my $count = (defined($1)) ? $2 : 1;  # defaults to 1
            die "Line $.: DC exceeds memory limit"
                if ($addr+$count-1 > 0xFFFF);
            #printf "DC:addr=%04X, count=%d\n", $addr, $count;
            AddUserDC($addr,$count);
            next;
        }

        if ($_ ne "") {
            print STDERR ">>$_<<";
            die "Line $.: unrecognized format";
        }

    } # while (<>)

    close(CFGFILE);
}

##########################################################################
# validate some of user's assertions

sub CheckConfig
{
    # see that all address user said were instructions have
    # been defined and are in fact legal instructions
    for(my $a=0x0000; $a<0xFFFF; $a += 2) {

        if (&hasAttr($a,"KNOWN_INST")) {
            if (!&valid16($a)) {
                printf "ERROR: User claims 0x%04X is an instruction, but memory is uninitialized\n", $a;
                exit(-1);
            }
            if ($decodeTable[$mem[$a]] eq "x") {
                printf "ERROR: User claims 0x%04X is an instruction, but it isn't legal\n", $a;
                exit(-1);
            }
        }

        if (&hasAttr($a,"KNOWN_INST") + &hasAttr($a,"KNOWN_DAC") + &hasAttr($a,"KNOWN_DC") > 1) {
            printf "ERROR: User claims multiple types at address 0x%04X\n", $a;
            exit(-1);
        }
    } # for
}

##########################################################################
# some utility functions

# verify that the address the user supplied is in range
sub checkAddr
{
    my($val,$linenum) = @_;
    if ($val > 0xFFFF) {
        printf "Address value of %04X on line %d is too big\n", $val, $linenum;
        exit(-1);
    }
}

# verify that the address the user supplied is in range and even
sub checkEvenAddr
{
    my($val,$linenum) = @_;
    &checkAddr($val,$linenum);
    if (($val & 1) == 1) {
        printf "Address value of %04X on line %d must be even\n", $val, $linenum;
        exit(-1);
    }
}

# returns true if the addressed location contains a valid instruction
sub IsValidInst
{
    my $addr = shift;
    return !($addr & 1) && &hasAttr($addr,["KNOWN_INST","INFR_INST"]);
}

# return the effective address of the instruction at the supplied address
sub EffAddr
{
    my $addr = shift;
    my $op   = $mem[$addr+0];
    my $imm  = $mem[$addr+1];
    my $dec  = $decodeTable[$op];
    my $ea = ($dec =~ "a") ? &AbsAddr($addr)            # absolute page 0/1
           : ($dec =~ "[cj]") ? (($addr & 0xFF00) | $imm)  # current page
           : 0;  # this shouldn't be used
}

# returns true if the instruction has a memory reference; else false
sub needsMref
{
    my $addr = shift;
    my $dec = $decodeTable[$mem[$addr]];
    return &IsValidInst($addr) && ($dec =~ m/[acj]/);
}

# if the instruction references memory, return the address if it can be
# determined, else return undef
sub getMref
{
    my $addr = shift;
    my $dec = $decodeTable[$mem[$addr]];
    if ($dec =~ m/a/) {
        # absolute addressing
        my $pstate = &getAttr($addr,"P_STATE");
        my $imm = $mem[$addr+1];
#printf "\n%04X, dec='$dec', abs, pstate='$pstate'\n", $addr;
        return ($pstate eq "0") ? (0x0000 + $imm) :
               ($pstate eq "1") ? (0x0100 + $imm) :
                                  undef;
    } elsif ($dec =~ m/[cj]/) {
        # current page addressing
#printf "%04X, dec='$dec', cur\n", $addr;
        my $imm = $mem[$addr+1];
        return ($addr & 0xFF00) | $imm;
    }
    # no memory reference
    return undef;
}

# returns true if the instruction has an indirect memory reference that
# might be constant; else false
sub needsIref
{
    my $addr = shift;
    my $dec = $decodeTable[$mem[$addr]];
    return &IsValidInst($addr) && ($dec =~ m/\* [ac]/);
}

# if the instruction makes an indirect reference to memory through a constant
# pointer, return the value of the pointer, otherwise return undef.
sub getIref
{
    my $addr = shift;

    my $mref = &getMref($addr);
    return undef if !defined($mref);

    # we don't count +/- since thing pointed to changes
    my $indir = ($decodeTable[$mem[$addr]] =~ m/[*]/);

    return undef
        if (!$indir || !valid16($mref) || IsModified($mref));

    return get16($mref);
}

# returns boolean indicating if byte pair is possibly written
sub IsWritten
{
    my $addr = shift;
    return &hasAttr($addr+0,["RMW2_SRC","WR2_SRC","RMW_SRC","WR_SRC"]) ||
           &hasAttr($addr+1,["RMW_SRC","WR_SRC"]);
}

# like IsWritten, but includes modification due to JST
sub IsModified
{
    my $addr = shift;
    return &hasAttr($addr,"JST_SRC") || IsWritten($addr);
}

# returns true if the page 0/page 1 status is known at location $addr
sub PageKnown
{
    my $addr = shift;
    return length(&getAttr($addr,"P_STATE")) == 1;
}

# get any references to this address
sub GetRefs
{
    my $addr = shift;
    return   &referer_list($addr, "BRANCH_SRC", "JMP")
           . &referer_list($addr, "JST_SRC",    "JST")
           . &referer_list($addr, "RD2_SRC",    "Rd2")
           . &referer_list($addr, "RD_SRC",     "Rd")
           . &referer_list($addr, "WR2_SRC",    "Wr2")
           . &referer_list($addr, "WR_SRC",     "Wr")
           . &referer_list($addr, "RMW2_SRC",   "Rmw2")
           . &referer_list($addr, "RMW_SRC",    "Rmw")
           . &referer_list($addr, "PTR_SRC",    "Ptr");
}

# pad spaces to end of reference string up to specified column
sub pad
{
    my($rstr,$col) = @_;
    my $pad = ($col-1) - length($$rstr);
    $pad = ($pad < 1) ? 1 : $pad;
    if ($pad < 1) {
        # make sure there is at least one space
        $$rstr .= " " if (substr($$rstr,-1) ne " ");
    } else {
        $$rstr .= " " x $pad;
    }
}


##########################################################################
# scan through memory and do some preprocessing of attributes that
# can be statically determined.

sub ClassifyInstructions
{
    for(my $a=0x0000; $a<0xFFFF; $a += 2) {
        if (&valid16($a)) {

            my $b0  = $mem[$a+0];
            my $b1  = $mem[$a+1];
            my $dec = $decodeTable[$b0];

            # memorize of the instruction is legal or not
            my $legal = ($dec ne "x") && !&hasAttr($a,["KNOWN_DAC","KNOWN_DC"]);
            setAttr($a,"LEGAL_INST", $legal);
            next if !$legal;

            # classify the instruction for subsequent processing
            my($branch, $memop, $pchange);
               if ($dec eq "JZA")                   { $branch  = "jza"   }
            elsif ($dec =~ /^J(GT|NE|LT|EQ)/)       { $branch  = "jcond" }
            elsif ($dec =~ /^J(MP|EI)/)             { $branch  = "jmp"   }
            elsif ($dec =~ /^JST/)                  { $branch  = "jst"   }
            elsif ($dec =~ /^S[TFAM][ASGI]/)        { $branch  = "skip"  }
            elsif ($dec =~ /j$/)                    { $branch  = "jmp"   }
            elsif ($dec =~ /^DU/)                   { $memop   = "WR2_SRC"  }
            elsif ($dec =~ /^U[AZ]/)                { $memop   = "WR_SRC"   }
            elsif ($dec =~ /^DAM/)                  { $memop   = "RMW2_SRC" }
            elsif ($dec =~ /^(AMC|INC|XMA)/)        { $memop   = "RMW_SRC"  }
            elsif ($dec =~ /^D(L|CM)/)              { $memop   = "RD2_SRC"  }
            elsif ($dec =~ / [ac]$/)                { $memop   = "RD_SRC"   }
            elsif ($dec =~ /^ONS/ && ( $b1 & 0x40)) { $pchange = "1" }
            elsif ($dec =~ /^OFS/ && (~$b1 & 0x40)) { $pchange = "0" }

            # this is special because it hits /j/ case too
               if ($dec =~ /^TAS/)                  { $pchange = "." }

            setAttr($a, "BRANCH",  $branch)  if defined($branch);
            setAttr($a, "MEMOP",   $memop)   if defined($memop);
            setAttr($a, "PCHANGE", $pchange) if defined($pchange);

        } # valid16
    } # for $a
}

##########################################################################
# scan through all of memory and at each point where the user declared
# an address as being legal, kick off a recursive search for more valid
# instructions starting at that point.

sub MarkInst
{
    my @opts = @_;

    %MarkOpt = ();
    $MarkOpt{$_} = 1 foreach (@opts);

    @Marked = ();

    $MarkPass++;
    if ($dbg > 0) {
        printf "======== Mark pass %d ========\n", $MarkPass;
        print  "Mark options for this pass: ", join(", ", @opts), "\n";
    }

    $MarkChanges = 0;

    for(my $a=0x0000; $a<0xFFFF; $a += 2) {
        if (&hasAttr($a,"KNOWN_INST")) {
            printf "%04X starts a new trace\n", $a if ($dbg > 1);
            &MarkRecursive($a, $a);
        }
    }

    print "This pass produced $MarkChanges updates\n" if ($dbg > 0);
}

##########################################################################
# recursively travel through memory, starting from a known good instruction
# and mark sequential instructions from that point as also being inferred
# instructions.  at each branch stop as we can't yet trust the indirect
# memory address as being the branch target.
#
# as we scan, if we hit an address that is already marked as an instruction
# then we return.
#
# as we scan, if we hit an illegal instruction, it puts into doubt whether
# that chain leading up to that bad instruction was indeed legal.  in the
# future we might do some backtracking, but for now, just return.

sub MarkRecursive
{
    my $addr  = shift;   # address of first instruction to inspect
    my $froma = shift;   # address that called this

    # this tracks what is known about the state of the P status bit
    # during an instruction trace.  it can be one of four values:
    #   "", "0", "1", "01"
    my $p_state = &getAttr($froma,"P_STATE");

    do {
        # force p_state as user wishes
        if (&hasAttr($addr,["KNOWN_P0","KNOWN_P1"])) {
            $p_state = (&getAttr($addr,"KNOWN_P0") ? "0" : "")
                     . (&getAttr($addr,"KNOWN_P1") ? "1" : "");
        }

        if ($dbg > 1) {
            printf "trying %04X", $addr;
            printf " from %04X", $froma if ($addr != $froma);
            print  "\n";
        }

        # there is no instruction here to process
        return if !valid16($addr) || !&getAttr($addr,"LEGAL_INST");

        if ($dbg > 1) {
            printf "     incoming P_STATE is '%s', currently '%s'\n", $p_state, &getAttr($addr,"P_STATE");
        }

        # stop if already inspected and the P_state hasn't changed
        my $pchanged = ($p_state ne &getAttr($addr,"P_STATE"));
        return if $Marked[$addr] && !$pchanged;

        # mark this address as having been traced, thus it is an instruction
        if (!$Marked[$addr]) {
            printf "     inferring instruction\n", $addr if ($dbg > 1);
            setAttr($addr, "INFR_INST");
        }

        if ($pchanged) {
            # merge the state that we thought we knew with the state
            # asserted by the approaching intruction stream
            printf "     revising P state to '%s'\n", $p_state
                if ($Marked[$addr] && ($dbg > 2));
            $p_state = (((&getAttr($addr,"P_STATE") . $p_state) =~ /0/) ? "0" : "")
                     . (((&getAttr($addr,"P_STATE") . $p_state) =~ /1/) ? "1" : "");
        }

        $Marked[$addr] = 1;

        # flags that classify instruction behavior
        my $branch  = &getAttr($addr,"BRANCH");  # type of branch
        my $memop   = &getAttr($addr,"MEMOP");   # memory operation
        my $pchange = &getAttr($addr,"PCHANGE"); # change in P flag

        my $op     = $mem[$addr+0];
        my $dec    = $decodeTable[$op];
        my $indir  = ($dec =~ m/[-+*]/);
        my $incdec = ($dec =~ m/[-+]/);

        # update the p-state tracking for this instruction
        if (&getAttr($addr,"P_STATE") ne $p_state) {
            printf "     P_STATE updated from '%s' to '%s'\n",
                   &getAttr($addr,"P_STATE"), $p_state if ($dbg > 2);
            setAttr($addr, "P_STATE", $p_state);
        }
        if (defined($pchange)) {
            $pchange =~ tr/.//d;
            $p_state = $pchange;
        }

        my $mref  = &getMref($addr);    # memory reference address (if any)
        my $iref  = &getIref($addr);    # post-indirection address, if constant
        my $finaddr = ($indir) ? $iref  # final address, no matter which mode
                               : $mref;

        if ($indir && defined($mref)) {
            setAttr($mref, "PTR_SRC",  $addr);
            setAttr($mref, "RMW2_SRC", $addr) if $incdec;
        }

        my @nextaddr;  # where to continue processing

        if ($branch eq "jza") {
            # we can't really track this one at all
            return;
        } elsif ($branch eq "jmp") {
            return if ($indir && !$MarkOpt{INDIR_JMP});
            if (defined($finaddr) && ($finaddr != $addr+2)) {
                setAttr($finaddr, "BRANCH_SRC", $addr);
            }
            push @nextaddr, $finaddr;
        } elsif ($branch eq "jcond") {
            if (defined($mref) && ($mref != $addr+2)) {
                setAttr($mref, "BRANCH_SRC", $addr);
                push @nextaddr, $mref;
            }
            push @nextaddr, $addr+2;
        } elsif ($branch eq "skip") {
            # FIXME: there is some risk here.  say we have
            #            STA X'??'  or   STA X'??'
            #            ONS @P          JST someplace
            #    the p_state might be affected by the skipped instruction,
            #    which should be reflected by the p_state we continue on with.
            @nextaddr = ( $addr+2, $addr+4 );
        } elsif ($branch eq "jst") {
            $finaddr = undef if ($indir && !$MarkOpt{INDIR_JST});
            my $rinfo;
            if (defined($finaddr)) {
                setAttr($finaddr, "JST_SRC", $addr);
                push @nextaddr, $finaddr + 2;  # continue at subroutine
                $rinfo = &getAttr($finaddr,"JST_INFO");
                if ($rinfo) {
                    # subroutine has unconventional return?
                    for my $i (0 .. $#{$rinfo}) {
                        my $ia = $addr + 2 + 2*$i;
                        my ($p,$c) = ($$rinfo[$i] =~ m/^(ret|param)(?:[:](.+))?$/);
                        $p =~ tr/A-Z/a-z/;
                        if ($p eq "param") {
                            # FIXME: this should be INFR_DAC, but we don't have one of those
                            #printf "Marking %04X as KNOWN_DAC\n", $addr +2 + 2*$i;
                            setAttr($ia, "KNOWN_DAC");
                        } elsif ($p eq "ret") {
                            #printf "Marking %04X as instruction\n", $addr +2 + 2*$i;
                            push @nextaddr, $ia;
                        } else {
                            die;  # should have been caught when parsing config
                        }
                        # FIXME: ideally, dasm3300_core would append the comment to its rv hash,
                        #        but that is awkward because it would require looking backwards
                        #        a few words to see if any of them have JST_INFO's that apply.
                        setAttr($ia, "COMMENTS", "; " . $c)
                            if ($c and !&hasAttr($ia, "COMMENTS"));
                    }
                } # rinfo
            }
            push @nextaddr, $addr+2 if (!$rinfo);  # conventional return
        } else {
            # not a branch op; sequential
            push @nextaddr, $addr+2;
        }

        # note memory references
        #printf "     fa=%04X, memop=$memop, addr=%04X\n", $finaddr, $addr if ($addr == 0x444A);
        setAttr($finaddr, $memop, $addr)
            if ($memop && !$incdec && defined($finaddr));

        # trace forward
        return if (!@nextaddr);
        if ($dbg > 2) {
            printf "     %04X: nextaddr list: ", $addr;
            print join(", ", map { sprintf "%04X", $_ } @nextaddr), "\n";
        }
        MarkRecursive( shift @nextaddr, $addr ) while ($#nextaddr > 0);

        # last one is handled with tail recursion
        $froma = $addr;
        $addr = shift @nextaddr;

    } while ($addr < 0xFFFF);
}

##########################################################################
# given a location in memory, analyze it and return a reference to a hash
# containing information on its subroutine attributes:
#
#    $$retval{SUB_OK}   is true if it is a well-formed subroutine
#    $$retval{SUB_LEAF} is true if this is a leaf subroutine
#    $$retval{SUB_LAST} contains the last return address
#
# A "well-formed" subroutine is one that
#     has clear return points
#     doesn't have any JMPs to it from outside it
#     doesn't have any JSTs to it from outside it (except at head)
#
# A leaf subroutine is one that is well-formed and doesn't call any other
# subroutines.

sub getSubroutineStatus
{
    my $addr    = shift;  # address to analyze
    my $verbose = shift;  # flag
    my %rv = ();          # return value

    return \%rv if (!&hasAttr($addr,"JST_SRC"));

    printf "Sub %04X:\n", $addr if $verbose;

    # make a list of addresses that JMP* to the subroutine link addr
    my @rets;
    if (&hasAttr($addr,"PTR_SRC")) {
        @rets = grep { $decodeTable[$mem[$_]] =~ m/^JMP\*/ }
                        @{&getAttr($addr,"PTR_SRC")};
        print  "    return points at ",
                join( ", ", map { sprintf("%04X", $_) }
                                  sort { $a <=> $b } @rets ),
                "\n" if $verbose;
    }

    # if there are no apparent return points, it isn't well formed
    return \%rv if (! @rets);

    # see if this subroutine is "well-formed" and perhaps a leaf

    my $wellformed = 1;
    my $leaf = 1;
    my $lastret = (sort { $a <=> $b} @rets)[-1];

    for(my $a=$addr+0; $a <= $lastret; $a += 2) {

        # check for jumps into it
        my $jmpsrcs = &getAttr($a, "BRANCH_SRC");
        if (defined($jmpsrcs)) {
            foreach my $src (@$jmpsrcs) {
                if ($src <= $addr or $src > $lastret) {
                    printf "    external jump from %04X to %04X\n",
                            $src, $a if $verbose;
                    $wellformed = 0;
                }
            }
        }

        # check for JSTs into it
        my $jstsrcs = &getAttr($a, "JST_SRC");
        if (($a > $addr) && defined($jstsrcs)) {
            printf "    jst to within subroutine, %04X to %04X\n",
                    $$jstsrcs[0],$a if $verbose;
            $wellformed = 0;
        }

        # check for branches outside
        my $br = &getAttr($a, "BRANCH");
        if ($br =~ /^(jcond|jmp|skip)/) {
            my $dec     = $decodeTable[$mem[$addr]];
            my $indir   = ($dec =~ m/[-+*]/);
            my $finaddr = ($indir) ? &getIref($addr)
                                   : &getMref($addr);
            if (defined($finaddr) and
                    $finaddr < $addr or $finaddr > $lastret) {
                printf "    routine makes external branch from %04X to %04X\n",
                        $a, $finaddr if $verbose;
                $leaf = 0
            }
        }

        # any JST means this isn't a leaf routine
        if ($br eq "jst") {
            printf "    routine makes subroutine call at %04X\n", $a
                                                                if $verbose;
            $leaf = 0;
        }
    }

    if ($wellformed && $verbose) {
        print "    This is a well-formed subroutine\n";
        print "    This is a leaf subroutine\n" if ($leaf);
    }

    $rv{SUB_OK}     = $wellformed;
    $rv{SUB_LEAF}   = $wellformed && $leaf;
    $rv{SUB_LAST}   = $lastret;
    return \%rv;
}


# return the P_STATE of a subroutine differs from its start P_STATE,
# return what the return value is, if it can be determined.
#     if all the return pstates are the same as the incoming pstate,
#         then this member returns a null
#     if all the return pstates concur and are different than the
#         incoming pstate, then this member holds that return state
#         ("0", "1", or "01").
#     if all the return pstates differ, then we note it with a warning
#         and return a null.  we could try to merge them, but not doing
#         so seems more conservative.
#     if no return points can be identified, null is also returned.
sub getSubReturnPstate
{
    my $addr    = shift;  # address to analyze
    my $verbose = shift;  # flag

    return undef if (!&hasAttr($addr,"JST_SRC"));

    return undef if !&hasAttr($addr,"PTR_SRC");

    # make a list of addresses that JMP* to the subroutine link addr
    my @rets = grep { $decodeTable[$mem[$_]] =~ m/^JMP\*/ }
                        @{&getAttr($addr,"PTR_SRC")};

    # if there are no apparent return points, it isn't well formed
    return undef if (! @rets);

    my $p_in  = &getAttr($addr+2,  "P_STATE");
    my $p_out = &getAttr($rets[0], "P_STATE");
    foreach my $reta (@rets[1..$#rets]) {
        my $p_this = &getAttr($reta, "P_STATE");
        if ($p_this ne $p_out) {
            printf "    inconsistent return P_STATES: '%s' (%04X) vs '%s' (%04X)\n",
                $p_in, $rets[0], $p_this, $reta
                    if $verbose;
            undef $p_out;
            last;
        }
    }
    if ($p_in ne $p_out) {
        print "    sub return P_STATE changed from '$p_in' to '$p_out'\n" if $verbose;
        return $p_out;
    }

    return undef;
}


##########################################################################
# scan through memory looking for subroutines and the corresponding
# return points.  we can't find them all, but most are discoverable.
# if we find a subroutine where the exit P_STATE isn't the same as the
# entrance P_STATE, then raise a flag.

sub AnalyzeSubroutines
{
    print "\n";
    print "==================== subroutine analysis ====================\n";

    for(my $a=0x0000; $a<0xFFFF; $a+=2) {
        &getSubroutineStatus($a, 1);
        &getSubReturnPstate($a, 1);
    }
}


##########################################################################
# generate labels as required for any memory references.
# there is one exception.  if this is an odd address of an instruction or
# a DAC and the even byte already has a label, then we don't make a label
# for this odd address.  any referers will refer to evenlabel+1.

sub CoinLabels
{
    for my $addr (0x0000 ... 0xFFFF) {
        if ( !&hasAttr($addr,"LABEL") &&
                (&hasAttr($addr,[ "BRANCH_SRC", "JST_SRC", "PTR_SRC",
                                  "RD2_SRC", "WR2_SRC", "RMW2_SRC",
                                  "RD_SRC", "WR_SRC", "RMW_SRC" ] )) ) {
            # no user supplied label, so make our own
            if (IsValidInst($addr-1)) {
                # FIXME: apply to any byte pair, not just instructoins?
                setAttr($addr-1, "LABEL", sprintf("L%04X", $addr-1))
                    if (!&hasAttr($addr-1,"LABEL"));
            } else {
                setAttr($addr, "LABEL", sprintf("L%04X", $addr));
            }
        }
    }
}

##########################################################################
# disassemble all of memory

sub Referenced
{
    my $addr = shift;
    return  defined $mem[$addr]
         || &hasAttr($addr,["LABEL", "PRECOMMENTS", "COMMENTS",
                            "KNOWN_DAC", "KNOWN_DC", "KNOWN_P0", "KNOWN_P1",
                            "BRANCH_SRC", "JST_SRC", "PTR_SRC",
                            "RD2_SRC", "WR2_SRC", "RMW2_SRC",
                            "RD_SRC", "WR_SRC", "RMW_SRC" ]);
}

sub DasmAll
{
    my $header = "Addr:";
    pad(\$header, $DATA_COLUMN);
    $header .= "Data";
    pad(\$header, $MREF_COLUMN);
    $header .= "Mref:Iref";
    pad(\$header, $TAG_COLUMN);
    $header .= "Tag";
    pad(\$header, $LABEL_COLUMN);
    $header .= "Label--";
    pad(\$header, $OPCODE_COLUMN);
    $header .= "Op--";
    pad(\$header, $ARG_COLUMN);
    $header .= "Arg------";
    pad(\$header, $COMMENT_COLUMN);
    $header .= "Comments----------------------";
    my $header2 = $header;
    $header  =~ s/-/ /g;        # kill hyphens
    $header2 =~ s/\S/=/g;       # convert to a bar
    print $header, "\n";
    print $header2, "\n";

    my $inst_len;
    for(my $a=0x0000; $a<=0xFFFF; $a += 0 + 1*$inst_len) {
        if (Referenced($a)) {
            my $rrv = dasm3300($a);
            $inst_len = $$rrv[0];
            foreach my $line (split "\n", $$rrv[1]) {
                print $line, "\n";
            }
        } else {
            $inst_len = 1;  # current byte undefined; try next one
        }
    }
}

##########################################################################
# print symbol table

sub PrintSymbolTable
{
    print "\n\n";
    print "Symbol table by address\n";
    print "=======================\n";

    for my $a (0x0000 ... 0xFFFF) {
        if (&hasAttr($a,"LABEL")) {
            printf "%04X %s\n", $a, &getAttr($a,"LABEL");
        }
    }
}

##########################################################################
# main routine

{
    my $filename = $ARGV[0];

    $dbg = 0;  # 0 is silent; 1, 2, 3 increasingly verbose

    &ProcessConfig($filename);
    &CheckConfig;
    &ClassifyInstructions;

    &MarkInst();             # most conservative
    &MarkInst("INDIR_JST");  # less conservative
    do {
        &MarkInst("INDIR_JST", "INDIR_JMP");  # still less conservative
    } while ($MarkChanges);

    &CoinLabels;
    #&AnalyzeSubroutines;
    &DasmAll;
    #&PrintSymbolTable;
}
